#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>

#include <stdlib.h>

#include "test_initng.h"

#include "../src/initng_load_module.h"

static void test_initng_load_module_init(void)
{
    initng_load_module_init();
    CU_ASSERT(g.modules == NULL);
    CU_ASSERT(strcmp(g.plugin_dir, INITNG_PLUGIN_DIR) == 0);
    initng_load_module_free_all();
    CU_ASSERT(g.plugin_dir == NULL);
}

static void test_initng_set_plugin_dir(void)
{
    initng_load_module_init();
    CU_ASSERT(strcmp(g.plugin_dir, INITNG_PLUGIN_DIR) == 0);
    initng_set_plugin_dir("/tmp");
    CU_ASSERT(strcmp(g.plugin_dir, "/tmp") == 0);
}

static void test_initng_get_next_module(void)
{
    m_h *m;

    /* check with no modules */
    m = initng_get_next_module(NULL);
    CU_ASSERT(m == NULL);

    /* XXX: check with non-empty module list */
}

static const char *get_module_path(const char *name)
{
    char dir[1024];
    static char full_name[1024];
    int n;

    /* get the full path to the module, so that dlopen() does not use 
     * the normal search algorithm */
    if (getcwd(dir, sizeof(dir)) == NULL) {
        fprintf(stderr, "%s:%d Error getting current dirctory\n", 
                __FILE__, __LINE__);
        exit(1);
    }
    n = snprintf(full_name, sizeof(full_name), "%s/plugins/%s", dir, name);
    if ((n < 0) || (n >= sizeof(full_name))) {
        fprintf(stderr, "%s:%d Module path too long\n", __FILE__, __LINE__);
        exit(1);
    }
    return full_name;
}

static void test_open_module(void)
{
    m_h m;
    const char *path;
    int ret;

    /* set our plugin directory */
    initng_set_plugin_dir("plugins");

    /* test a non-existant plugin */
    path = get_module_path("libplug_nonexistant.so");
    ret = open_module(path, "nonexistant", &m);
    CU_ASSERT(ret == FALSE);

    /* test the simplest plugin */
    path = get_module_path("libplug_empty.so");
    ret = open_module(path, "plug_empty", &m);
    CU_ASSERT(ret == TRUE);
    if (ret) {
        CU_ASSERT(strcmp(m.module_name, "plug_empty") == 0);
        CU_ASSERT(m.module_dlhandle != NULL);
        CU_ASSERT(m.module_init != NULL);
        CU_ASSERT(m.module_unload != NULL);
        CU_ASSERT(m.module_needs == NULL);
        close_module(&m);
    }

    /* missing module_init() */
    path = get_module_path("libplug_broken_0.so");
    ret = open_module(path, "plug_broken_0", &m);
    CU_ASSERT(ret == FALSE);

    /* missing module_unload() */
    path = get_module_path("libplug_broken_1.so");
    ret = open_module(path, "plug_broken_1", &m);
    CU_ASSERT(ret == FALSE);

    /* module_needs specified, but empty */
    path = get_module_path("libplug_needs.so");
    ret = open_module(path, "plug_needs", &m);
    CU_ASSERT(ret == TRUE);
    if (ret) {
        CU_ASSERT(strcmp(m.module_name, "plug_needs") == 0);
        CU_ASSERT(m.module_dlhandle != NULL);
        CU_ASSERT(m.module_init != NULL);
        CU_ASSERT(m.module_unload != NULL);
        CU_ASSERT(m.module_needs != NULL);
        CU_ASSERT(m.module_needs[0] == NULL);
        close_module(&m);
    }

    /* module_needs specified, non-empty */
    path = get_module_path("libplug_needs_empty.so");
    ret = open_module(path, "plug_needs_empty", &m);
    CU_ASSERT(ret == TRUE);
    if (ret) {
        CU_ASSERT(strcmp(m.module_name, "plug_needs_empty") == 0);
        CU_ASSERT(m.module_dlhandle != NULL);
        CU_ASSERT(m.module_init != NULL);
        CU_ASSERT(m.module_unload != NULL);
        CU_ASSERT(m.module_needs != NULL);
        CU_ASSERT(strcmp(m.module_needs[0], "plug_empty") == 0);
        CU_ASSERT(m.module_needs[1] == NULL);
        close_module(&m);
    }
}

static void test_initng_load_module(void)
{
    int ret;
    m_h *m;

    /* set our plugin directory */
    initng_set_plugin_dir("plugins");

    /* verify we start empty */
    m = initng_get_next_module(NULL);
    CU_ASSERT(m == NULL);

    /* test the simplest plugin */
    ret = initng_load_module("plug_empty");
    CU_ASSERT(ret == TRUE);
    m = initng_get_next_module(NULL);
    CU_ASSERT(m != NULL);
    CU_ASSERT(strcmp(m->module_name, "plug_empty") == 0);

    /* verify we can't load it twice */
    ret = initng_load_module("plug_empty");
    CU_ASSERT(ret == FALSE);

    /* try loading a bogus module */
    ret = initng_load_module("plug_nonexistant");
    CU_ASSERT(ret == FALSE);

    /* verify that we can load a module that has all needs */
    ret = initng_load_module("plug_needs_empty");
    CU_ASSERT(ret == TRUE);
/* XXX: check state of world here */

    /* unload the loaded modules */
    initng_unload_module_named("plug_needs_empty"); 
    initng_unload_module_named("plug_empty"); 

    /* verify that we can't load a module that does *NOT* have all it needs */
    ret = initng_load_module("plug_needs_empty");
    CU_ASSERT(ret == FALSE);
}

static void test_initng_unload_module_named(void)
{
    m_h *m;

    /* verify we start empty */
    m = initng_get_next_module(NULL);
    CU_ASSERT(m == NULL);

    /* try unloaded a module that is not loaded */
    CU_ASSERT(initng_unload_module_named("not_loaded") == FALSE);

    /* try a load and unload */
    CU_ASSERT(initng_load_module("plug_empty") == TRUE);
    CU_ASSERT(initng_unload_module_named("plug_empty") == TRUE);

    /* verify we can't unload again */
    CU_ASSERT(initng_unload_module_named("plug_empty") == FALSE);

    /* try to unload a needed module */
    CU_ASSERT(initng_load_module("plug_empty") == TRUE);
    CU_ASSERT(initng_load_module("plug_needs_empty") == TRUE);
    CU_ASSERT(initng_unload_module_named("plug_empty") == FALSE);

    /* verify we can unload in the proper order */
    CU_ASSERT(initng_unload_module_named("plug_needs_empty") == TRUE);
    CU_ASSERT(initng_unload_module_named("plug_empty") == TRUE);
}

static void test_initng_load_all_modules(void)
{
    m_h *m;

    /* set our plugin directory */
    initng_set_plugin_dir("plugins");

    /* load the modules */
    CU_ASSERT(initng_load_all_modules());

    /* 
     * our list should be one of:
     *  needs_empty -> needs -> empty
     *  needs_empty -> empty -> needs
     *  needs -> needs_empty -> empty
     */
    m = initng_get_next_module(NULL);
    CU_ASSERT(!strcmp(m->module_name, "plug_needs_empty") || 
              !strcmp(m->module_name, "plug_needs"));
    if (!strcmp(m->module_name, "plug_needs_empty")) {
        m = initng_get_next_module(m);
        CU_ASSERT(!strcmp(m->module_name, "plug_needs") || 
                  !strcmp(m->module_name, "plug_empty"));
        m = initng_get_next_module(m);
        CU_ASSERT(!strcmp(m->module_name, "plug_needs") || 
                  !strcmp(m->module_name, "plug_empty"));
    } else {
        m = initng_get_next_module(m);
        CU_ASSERT(!strcmp(m->module_name, "plug_needs_empty"));
        m = initng_get_next_module(m);
        CU_ASSERT(!strcmp(m->module_name, "plug_empty"));
    }
    m = initng_get_next_module(m);
    CU_ASSERT(m == NULL);

    /* unload the modules */
    initng_unload_all_modules();

    /* should be empty */
    m = initng_get_next_module(m);
    CU_ASSERT(m == NULL);
}

static void test_initng_load_module_free_all(void)
{
    initng_load_module_free_all();
    CU_ASSERT(g.plugin_dir == NULL);
}

void test_initng_load_module_setup(void)
{
    CU_pSuite suite;

    suite = CU_add_suite("initng_load_module", empty_init, empty_cleanup);
    CU_add_test(suite, "initng_load_module_init", test_initng_load_module_init);
    CU_add_test(suite, "initng_set_plugin_dir", test_initng_set_plugin_dir);
    CU_add_test(suite, "test_initng_get_next_module", 
                    test_initng_get_next_module);
    CU_add_test(suite, "open_module", test_open_module);
    CU_add_test(suite, "initng_load_module", test_initng_load_module);
    CU_add_test(suite, "initng_unload_module_named", 
                    test_initng_unload_module_named);
    CU_add_test(suite, "initng_load_all_modules", test_initng_load_all_modules);
    CU_add_test(suite, "initng_load_module_free_all", 
                    test_initng_load_module_free_all);
}
